//CN-Palm Summer 2003
//read(socket,buffer,buflen)
//write(socket,buf,buflen)
//07/10/03 Added TCPReadLine()
//         Added #defines for unix/palm socket API
//10/10/03 Haha, forgot to decrease buffer counter, no wonder it
//          appeared to pack up after some long ?help messages.

#include "net.h"

static void TCPShowError(Err err);

static NetSocketRef socket = -1;
#ifndef EASY
static NetSocketAddrMyType sockAddr;
#endif

static Char host[16];
static UInt16 port;

static Char recvBuf[BUF_SIZE];
static Int16 received;

void TCPReceive()
{
	Err err;
	Int16 read=-1, spaceleft;

#ifndef EASY
	NetSocketAddrType from;
	UInt16 size = sizeof(NetSocketAddrMyType);
#endif

	if (!bConnected) return;
	spaceleft = BUF_SIZE - received;
	MemSet(&recvBuf[received], spaceleft, 0);

#ifdef EASY
	read = recv(socket, recvBuf, spaceleft, 0);
	err = errno;
#else
	//from and size can be NULL
	read = NetLibReceive(AppNetRefnum, socket,
		&recvBuf[received], spaceleft, 0, &from, &size,
		sysTicksPerSecond, &err);
#endif

	switch(read) {
	case 0:
//		DAddMessage("disconnected by server(?)");
//		TCPDisconnect();
		TCPShowError(err);
		break;
	case -1:
		//if (err) DAddMessage("receive error");
		if (err == netErrTimeout)
			break;
		TCPShowError(err);
		break;
	default:
		received = received + read;
//		DAddMessage(recvBuf);
/*
		{
			Char str[16];
			StrPrintF(str, "%d (%d) %db",
				sizeof(NetSocketAddrMyType), size, received);
			DAddMessage(str);
		}
*/
		break;
	}
}

Char *TCPReadLine()
{
	Int16 i=BUF_SIZE;
	Char *line;

	if (received < 1) return NULL;

	for(i=0; i<received; i++)
	{
		if (recvBuf[i] == '\r' || recvBuf[i] == '\n')
			break;
	}
	if (i >= received) return NULL;

	recvBuf[i++] = '\0';
	line = MemPtrNew(i);
	MemMove(line, recvBuf, i);

	//CR and LF removal
	if (i < received)
	{
		if (recvBuf[i] == '\r' || recvBuf[i] == '\n')
			i++;
	}

	//Decrease counter (v. important)
	received = received - i;

	MemMove(recvBuf, &recvBuf[i], BUF_SIZE-i);

	return line;
}

//Returns 1 on success
Int8 TCPSendRaw(Char *buf, UInt16 len)
{
	Err err;
	if (!bConnected) return 0;

#ifdef EASY
	send(socket, buf,len, 0); err = errno;
#else
	NetLibSend(AppNetRefnum, socket, buf, len, 0,0,0,
		sysTicksPerSecond*2, &err);
#endif
	TCPShowError(errno);

	return err ? 0 : 1;
}

//Returns 1 on success
Int8 TCPConnect()
{
#ifndef EASY
	Err err;
#endif
	if (bConnected) return 1;
	DAddMessage("Connecting...");

#ifdef EASY
	//Palm Inc does not recommend this method
	DAddMessage(host);
	socket = NetUTCPOpen(host, NULL, port);
	if (socket < 0)
	{
		DAddMessage("NetUTCPFailed");
		return 0;
	}
#else
	socket = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET,
		netSocketTypeStream, netSocketProtoIPTCP, AppNetTimeout,
		&err);
	if (err || socket < 0)
	{
		DAddMessage("Could not open socket:");
		TCPShowError(err);
		return 0;
	}

	sockAddr.family = netSocketAddrINET;
	NetLibSocketConnect(AppNetRefnum, socket,
		(NetSocketAddrType*)&sockAddr,
		sizeof(NetSocketAddrMyType), AppNetTimeout, &err);
	if (err)
	{
		DAddMessage("Could not connect socket:");
		TCPShowError(err);
		return 0;
	}
#endif

	bConnected = 1;
	return 1;
}

//Returns 1 on success
Int8 TCPDisconnect()
{
	Err err;
	if (!bConnected) return 1;
	DAddMessage("Disconnecting...");

	if (socket >= 0)
	{
		NetLibSocketClose(AppNetRefnum, socket, 2000, &err);
		TCPShowError(err);
	}

	bConnected = 0;
	return err ? 0 : 1;
}

//// Util ////
void TCPInit()
{
	Int16 netIFErrs;
	Err err = SysLibFind("Net.lib", &AppNetRefnum);
	if (err) DAddMessage("Cannot find Net Library");

	err = NetLibOpen(AppNetRefnum, &netIFErrs);
	TCPShowError(err);
	AppNetTimeout = 2 * sysTicksPerSecond;
	received = 0;
}

void TCPEnd()
{
	Err err = NetLibClose(AppNetRefnum, true);
	if (err == netErrStillOpen)
		DAddMessage("NetLib left open (OK)");
}

#ifdef EASY
void TCPSetHost(Char *host_, UInt16 port_)
{
	StrCopy(host, host_);
	port = port_;
}
#else
void TCPSetHost(UInt8 a, UInt8 b, UInt8 c, UInt8 d, UInt16 port)
{
	sockAddr.addr.part.a = a;
	sockAddr.addr.part.b = b;
	sockAddr.addr.part.c = c;
	sockAddr.addr.part.d = d;
	sockAddr.port        = port;
}
#endif

static void TCPShowError(Err err) {
	switch(err) {
	case 0: break;
	case netErrAlreadyOpen:
		DAddMessage("NetLib already open");
		break;
	case netErrNotOpen:
		DAddMessage("NetLib not open");
		break;
	case netErrParamErr:
		DAddMessage("Bad code (params)");
		break;
	case netErrNoMoreSockets:
		DAddMessage("No sockets left :(");
		break;
	case netErrOutOfMemory:
		DAddMessage("Out of memory");
		break;
	case netErrSocketNotOpen:
		DAddMessage("Socket not open");
		break;
	case netErrNoInterfaces:
		DAddMessage("Setup Prefs first"); break;
	case netErrTimeout:
		DAddMessage("Socket timed out");
		break;
	case netErrOutOfCmdBlocks:
		DAddMessage("Out of \"CmdBlocks\"!?");
		break;
	case netErrPrefNotFound:
		DAddMessage("Setup Prefs first");
		break;

	case netErrUnreachableDest:
		DAddMessage("Unreachable destination");
	case 4628:
		DAddMessage("Connection broken");
		TCPDisconnect();
		break;
	default:
//4628
		{
			Char str[32];
			StrPrintF(str, "net.c: Error: %d", err);
			DAddMessage(str);
		}
		break;
	}
}
